/* Copyright 2011 Jetro Lauha - http://iki.fi/jetro/
 *
 * $Id: MainState.cpp 371 2011-12-18 22:59:30Z tonic $
 * $Revision: 371 $
 */

#include <time.h>
#include <assert.h>
#include <algorithm>

#include "main.h"
#include "MainState.h"


#define PLR_SPEED               1.0f
#define PLR_IDLEFRAMES          30

#define PLR_W_PX                11
#define PLR_H_PX                21


#define MUSIC_VOLUME    70
#define AUDIO_CHANNELS  8

static FSOUND_STREAM *sMusicStream;
static int sMusicChannel = -1;
static float sMusicFadeIn = 0;


static float sClamp(float value, float min, float max)
{
    return value < min ? min : (value > max ? max : value);
}

static float sStep(float value, float stepPosition)
{
    return (float)(value < stepPosition ? 0 : 1);
}

static float sBoxStep(float value, float slopeStart, float slopeEnd)
{
    float diff = slopeEnd - slopeStart;
    if (diff == 0)
        return sStep(value, slopeStart);
    return sClamp((value - slopeStart) / diff, 0, 1);
}


void gIrrDrawFullscreenQuad(video::IVideoDriver *driver, video::SColor color, const video::SMaterial *material, const core::vector2df *uvs)
{
    video::SMaterial mat;
    if (material)
        mat = *material;
    else
    {
        if (color.getAlpha() != 0xff)
        {
            mat.MaterialType = video::EMT_ONETEXTURE_BLEND;
            mat.MaterialTypeParam = video::pack_texureBlendFunc(video::EBF_SRC_ALPHA, video::EBF_ONE_MINUS_SRC_ALPHA, video::EMFN_MODULATE_1X, video::EAS_VERTEX_COLOR);
        }
        mat.setFlag(video::EMF_LIGHTING, false);
        mat.setFlag(video::EMF_ZWRITE_ENABLE, false);
        mat.ColorMaterial = video::ECM_DIFFUSE_AND_AMBIENT;
        mat.ZBuffer = video::ECFN_NEVER;
        mat.DiffuseColor = color;
    }
    driver->setMaterial(mat);
    video::S3DVertex vertices[4];
    for (int a = 0; a < 4; ++a)
        vertices[a].Color = color;
    const core::dimension2du &screenSize = driver->getScreenSize();
    vertices[0].Pos.set(0, 0, 0);
    vertices[1].Pos.set((f32)screenSize.Width, 0, 0);
    vertices[2].Pos.set(0, (f32)screenSize.Height, 0);
    vertices[3].Pos.set((f32)screenSize.Width, (f32)screenSize.Height, 0);
    if (uvs)
    {
        vertices[0].TCoords.set(uvs[0].X, uvs[0].Y);
        vertices[1].TCoords.set(uvs[1].X, uvs[0].Y);
        vertices[2].TCoords.set(uvs[0].X, uvs[1].Y);
        vertices[3].TCoords.set(uvs[1].X, uvs[1].Y);
    }
    const u16 indices[6] = { 0, 1, 2,  3, 2, 1 };
    driver->drawVertexPrimitiveList(vertices, 4, indices, 2, video::EVT_STANDARD, scene::EPT_TRIANGLES, video::EIT_16BIT);
}


MainState::MainState()
{
    for (int a = 0; a < K_TYPE_COUNT; ++a)
        mKeyState[a] = false;
}


MainState::~MainState()
{
}


bool MainState::init()
{
    Graphics *g = App::getInstance()->getGraphics();

    if (!FSOUND_Init(44100, AUDIO_CHANNELS, FSOUND_INIT_DONTLATENCYADJUST))
    {
        gErrorMessage("Audio error", "Unable to initialize sound! (%d)",
                      FSOUND_GetError());
    }
    sMusicStream = FSOUND_Stream_Open("data/music.ogg", FSOUND_LOOP_NORMAL, 0, 0);
    if (!sMusicStream)
        gErrorMessage("Audio error", "Can't load music");
    else
    {
        // set stream start time by current time
        int streamLengthBytes = FSOUND_Stream_GetLength(sMusicStream);
        int streamLengthMs = FSOUND_Stream_GetLengthMs(sMusicStream);
        int streamLengthSecs = streamLengthMs / 1000;
        int t = (int)time(NULL);
        int startPos = (t % streamLengthSecs) * 1000;
        int startPosBytes = (int)(startPos * ((double)streamLengthBytes / streamLengthMs));
        FSOUND_Stream_SetPosition(sMusicStream, startPosBytes);
        sMusicChannel = FSOUND_Stream_Play(FSOUND_FREE, sMusicStream);
        FSOUND_SetVolume(sMusicChannel, 0);
        sMusicFadeIn = 0;
    }

    mFont = new BMFont(g, "data/terminal.fnt");
    if (!mFont->init())
        return false;

    // Make sure these are in synch with IMAGE_ID
    mImages.push_back(g->loadImage("data/button.png"));
    mImages.push_back(g->loadImage("data/button-hot.png"));
    mImages.push_back(g->loadImage("data/button-active.png"));
    mImages.push_back(g->loadImage("data/button-hot-active.png"));
    mImages.push_back(g->loadImage("data/sprites.png"));
    mImages.push_back(g->loadImage("data/stage-baby_png/heart.png")); // the heart gets loaded more than one time, in multi-copies.. let's just use one of those

    mButtonStyle = new IMGUIImageButtonStyle(mFont, mImages[IMG_BUTTON], mImages[IMG_BUTTON_HOT], mImages[IMG_BUTTON_ACTIVE], mImages[IMG_BUTTON_HOT_ACTIVE]);

    mStages.setMainState(this);
    bool stageGfxOk = mStages.loadStageGfxLayers(g);
    if (!stageGfxOk)
        return false;

    mDevice = App::getInstance()->getIrrDevice();
    mVideo = App::getInstance()->getIrrVideoDriver();

    mLastMousePos.set(WIDTH / 2, HEIGHT / 2);

    mStages.setStage(STAGE_BABY);

    mEndingUIActive = false;
    mTitleUIActive = true;
    mInGame = false;
    srand((unsigned int)time(NULL));
    //startGame(); // testing

    return mFont != NULL && mButtonStyle != NULL;
}


void MainState::deinit()
{
    UI32 a;

    mStages.freeStageGfxLayers();

    delete mFont;
    for (a = 0; a < mImages.size(); ++a)
        delete mImages[a];
    mImages.clear();

    FSOUND_Stream_Stop(sMusicStream);
    FSOUND_Stream_Close(sMusicStream);
    for (a = 0; a < AUDIO_CHANNELS; ++a)
        FSOUND_StopSound(a);
    FSOUND_Close();
}


void MainState::onActivate(UI32 time)
{
    AppState::onActivate();
    mPaused = false;
}


void MainState::startGame()
{
    mGameLogicTime = 0;

    mScore = 0;
    mPlayer.state = PS_STILL;
    int px, py;
    mStages.getStageSpawn(px, py);
    mPlayer.x = mPlayer.prevX = (float)px;
    mPlayer.y = mPlayer.prevY = (float)py;

    mTitleUIActive = false;
    mInGame = true;

    mStageChangeScheduled = false;
    mFadePhase = -1.0f;
}


bool MainState::onKeyUp(const irr::SEvent::SKeyInput &event)
{
    if (event.Key == irr::KEY_LEFT)
        mKeyState[K_LEFT] = false;
    else if (event.Key == irr::KEY_RIGHT)
        mKeyState[K_RIGHT] = false;
    else if (event.Key == irr::KEY_UP)
        mKeyState[K_UP] = false;
    else if (event.Key == irr::KEY_DOWN)
        mKeyState[K_DOWN] = false;
    else if (event.Key == irr::KEY_KEY_Z)
        mKeyState[K_ACTION] = false;

    return false;
}

bool MainState::onKeyDown(const irr::SEvent::SKeyInput &event)
{
    if (event.Key == irr::KEY_ESCAPE)
        App::getInstance()->setAlive(false);

    if (event.Key == irr::KEY_LEFT)
        mKeyState[K_LEFT] = mKeyState[K_FLAG_LEFT] = true;
    else if (event.Key == irr::KEY_RIGHT)
        mKeyState[K_RIGHT] = mKeyState[K_FLAG_RIGHT] = true;
    else if (event.Key == irr::KEY_UP)
        mKeyState[K_UP] = mKeyState[K_FLAG_UP] = true;
    else if (event.Key == irr::KEY_DOWN)
        mKeyState[K_DOWN] = mKeyState[K_FLAG_DOWN] = true;
    else if (event.Key == irr::KEY_KEY_Z)
        mKeyState[K_ACTION] = mKeyState[K_FLAG_ACTION] = true;

    return false;
}

bool MainState::onMouseDown(const irr::SEvent::SMouseInput & /*event*/)
{
    if (mTitleUIActive || mEndingUIActive)
        mIMGUI.onMouseDown();

    return false;
}

bool MainState::onMouseUp(const irr::SEvent::SMouseInput & /*event*/)
{
    mIMGUI.onMouseUp();
    return false;
}

bool MainState::onMouseMotion(const irr::SEvent::SMouseInput &event)
{
    mLastMousePos.set(event.X, event.Y);
    mIMGUI.onMouseMotion(event.X, event.Y);
    return false;
}

bool MainState::onMouseWheel(const irr::SEvent::SMouseInput & /*event*/)
{
    return false;
}


void MainState::gameFinished()
{
    //mInGame = false;
    mEndingUIActive = true;
}


bool MainState::update(I32 deltaTime)
{
    sMusicFadeIn += 0.005f;
    if (sMusicFadeIn > 1)
        sMusicFadeIn = 1;
    if (sMusicStream && sMusicChannel >= 0)
        FSOUND_SetVolume(sMusicChannel, (int)(sMusicFadeIn * MUSIC_VOLUME));

    bool result = AppState::update(deltaTime);
    if (!result)
        return false; // game is paused...

    if (mTitleUIActive)
    {
        if (mKeyState[K_ACTION] || mKeyState[K_FLAG_ACTION])
        {
            startGame();
        }
        return false;
    }

    // Following is executed only when game is not paused & game is running.

    float plrSpeed = PLR_SPEED;

    // hack..
    if (mStages.getCurrentStage() == STAGE_HOMEOLD)
        plrSpeed *= 0.3f;

    mPlayer.state = PS_STILL;
    float orgPlrX = mPlayer.x;
    float orgPlrY = mPlayer.y;
    if (mKeyState[K_LEFT] || mKeyState[K_FLAG_LEFT])
    {
        mPlayer.x -= plrSpeed;
        mPlayer.spriteDir = PD_LEFT;
        mPlayer.state = PS_MOVING;
    }
    if (mKeyState[K_RIGHT] || mKeyState[K_FLAG_RIGHT])
    {
        mPlayer.x += plrSpeed;
        mPlayer.spriteDir = PD_RIGHT;
        mPlayer.state = PS_MOVING;
        if (mKeyState[K_LEFT] || mKeyState[K_FLAG_LEFT])
            mPlayer.spriteDir = PD_DOWN; // look down if both left+right held down
    }
    if (mKeyState[K_UP] || mKeyState[K_FLAG_UP])
    {
        mPlayer.y -= plrSpeed;
        mPlayer.spriteDir = PD_UP;
        mPlayer.state = PS_MOVING;
    }
    if (mKeyState[K_DOWN] || mKeyState[K_FLAG_DOWN])
    {
        mPlayer.y += plrSpeed;
        mPlayer.spriteDir = PD_DOWN;
        mPlayer.state = PS_MOVING;
    }
    mPlayer.prevY = orgPlrY;
    mPlayer.prevX = orgPlrX;

    int levelW, levelH;
    unsigned int *collData;
    bool stageHasCollisions = mStages.getStageCollisionMetaData(levelW, levelH, collData);
    bool metaDataCollState[SMCI_COUNT];
    for (int a = 0; a < SMCI_COUNT; ++a)
        metaDataCollState[a] = false;

    if (stageHasCollisions && (mStages.getStageCtrlMask() & STAGE_CTRL_MOVE))
    {
        assert(collData);
        assert(levelW > 0 && levelH > 0);

        int ptx = (int)floorf(mPlayer.x);
        int pty = (int)floorf(mPlayer.y);

        bool collision = false;
        // HACK!!! this is not a good way, but had to get it work somehow
        // this way guy gets stuck instead of sliding :(, but I was in a real hurry

        const int px0 = (int)mPlayer.x - PLR_W_PX / 2;
        const int px1 = px0 + PLR_W_PX;
        const int py0 = (int)mPlayer.y - PLR_H_PX / 2;
        const int py1 = py0 + PLR_H_PX;

        for (int tly = py0; tly < py1; ++tly)
        {
            if (tly < 0 || tly >= levelH)
                continue;
            for (int tlx = px0; tlx < px1; ++tlx)
            {
                if (tlx < 0 || tlx >= levelW)
                    continue;
                unsigned int collPx = collData[tly * levelW + tlx];
                unsigned int alpha = (unsigned int)(collPx & SMC_MASK_ALPHA);
                unsigned int alphaTst = (unsigned int)SMC_MASK_ALPHA / 2;
                if (alpha < alphaTst) // ignore if too transparent
                    continue;
                collPx &= SMC_MASK_COLOR;

                for (int a = 0; a < SMCI_COUNT; ++a)
                {
                    if (collPx == sStageMetaDataCollValues[a])
                        metaDataCollState[a] = true;
                }

                if (collPx != sStageMetaDataCollValues[SMCI_WALL])
                    continue; // this px is not a wall

                collision = true;
            } // tlx
        } // tly

        if (collision)
        {
            mPlayer.x = mPlayer.prevX;
            mPlayer.y = mPlayer.prevY;
        }
    } // stageHasCollisions

    bool stageDone = mStages.update(mStateTime, mKeyState, metaDataCollState);

    for (int a = 0; a < K_TYPE_COUNT; ++a)
    {
        if (mKeyState[a])
            mPlayer.idleCount = 0;
    }
    mKeyState[K_FLAG_LEFT] = mKeyState[K_FLAG_RIGHT] = mKeyState[K_FLAG_UP] = mKeyState[K_FLAG_DOWN] = mKeyState[K_FLAG_ACTION] = false;

    if (++mPlayer.idleCount >= PLR_IDLEFRAMES && mPlayer.state == PS_STILL)
        mPlayer.state = PS_IDLE;

    if (stageDone)
    {
        mStageChangeScheduled = true;
    }

    if (mFadePhase < 0)
        mFadePhase += 0.01f;
    else if (mStageChangeScheduled && mFadePhase < 1.0f)
    {
        mFadePhase += 0.01f;
        if (mFadePhase >= 1.0f)
        {
            bool moreStatesToGo = mStages.getCurrentStage() + 1 < mStages.getStageCount();
            if (moreStatesToGo)
            {
                mStages.setStage(mStages.getCurrentStage() + 1);
                mFadePhase = -1.0f;
                mStageChangeScheduled = false;

                int levelW, levelH;
                unsigned int *collData;
                bool stageHasCollisions = mStages.getStageCollisionMetaData(levelW, levelH, collData);
                if (stageHasCollisions)
                {
                    // new spawn
                    int px, py;
                    mStages.getStageSpawn(px, py);
                    mPlayer.x = mPlayer.prevX = (float)px;
                    mPlayer.y = mPlayer.prevY = (float)py;
                }
            }
            else
                App::getInstance()->setAlive(false); // NOTE: QUITS
        }
    }
    else
        mFadePhase = 0;

    return true;
}


void MainState::render(irr::IrrlichtDevice *device)
{
    assert(device);
    irr::video::IVideoDriver *driver = device->getVideoDriver();
    assert(driver);

    driver->beginScene(true, true, video::SColor(0));

    Graphics *g = App::getInstance()->getGraphics();
    g->beginRender();

    if (mInGame)
    {
        renderGame(device);
    }
    else
    {
        mIMGUI.begin(mStateTime);

        if (mTitleUIActive)
        {
            mFont->drawString(WIDTH / 2, 120, "LIFELONE", IF_ALIGN_HORIZ_CENTER, 1, Color(0xffffffff));

            mFont->drawText(WIDTH / 2, 200, "A game for 22th Ludum Dare 48h game compo\n\nTheme: Alone", IF_ALIGN_TOP_HCENTER, 1, 0, Color(0xff888888));

            if (mIMGUI.button(WIDTH / 2 - mButtonStyle->getWidth() / 2, 300, IMGUI_AI_COUNTER, mButtonStyle, 0, 0, "Play"))
            {
                startGame();
            }

            mFont->drawString(WIDTH / 2, HEIGHT - 2, "Copyright 2011 Jetro Lauha - www.jet.ro", IF_ALIGN_BOTTOM_HCENTER, 1, Color(0xff888888));
        }

        mIMGUI.end();
    }

    g->endRender();

    driver->endScene();
}


void MainState::drawSprite(float x, float y, int num, bool horizFlip)
{
    static Image *spriteImg = mImages[IMG_SPRITES];
    static const int spriteSize = 24;
    static const int spritesPerRow = (int)(spriteImg->getWidth() / spriteSize);
    int col = num % spritesPerRow;
    int row = num / spritesPerRow;
    Rectangle srcRect((float)(col * spriteSize), (float)(row * spriteSize), (float)spriteSize, (float)spriteSize);
    spriteImg->draw(IDM_NORMAL, IF_ALIGN_CENTER | (horizFlip ? IF_FLIP_HORIZ : 0), x, y, Color(0xffffffff), &srcRect);
}

void MainState::drawPlayerSprite(float x, float y, STAGE_PLR_TYPE type, int num, bool horizFlip)
{
    static Image *spriteImg = mImages[IMG_SPRITES];
    static const int spriteSize = 24;
    static const int spritesPerRow = 4;
    int col = num % spritesPerRow;
    int row = num / spritesPerRow;
    col += type * spritesPerRow; // select 4x4 group horizontally by type
    Rectangle srcRect((float)(col * spriteSize), (float)(row * spriteSize), (float)spriteSize, (float)spriteSize);
    spriteImg->draw(IDM_NORMAL, IF_ALIGN_CENTER | (horizFlip ? IF_FLIP_HORIZ : 0), floorf(x), floorf(y), Color(0xffffffff), &srcRect);
}


void MainState::renderGame(irr::IrrlichtDevice *device)
{
    irr::video::IVideoDriver *driver = device->getVideoDriver();
    App *app = App::getInstance();
    Graphics *g = app->getGraphics();

    // TODO/FIXME: draw bkg stuff here and/or replace the tmp code below
/*
    g->endRender(); //TMP - render() already called beginRender but we use some base irr drawrects below temporarily so we need to cancel it.
    const float scale = 64.f;
    const float scx = app->getScreenWidth() / 2.0f, scy = app->getScreenHeight() / 2.0f;
    float cx = mPlayer.x, cy = mPlayer.y;
    int tileVisTop = core::clamp<int>((int)(cy - scy / scale), 0, mLevelHeight - 1);
    int tileVisBottom = core::clamp<int>((int)(cy + scy / scale + 1), 0, mLevelHeight - 1);
    int tileVisLeft = core::clamp<int>((int)(cx - scx / scale), 0, mLevelWidth - 1);
    int tileVisRight = core::clamp<int>((int)(cx + scx / scale + 1), 0, mLevelWidth - 1);
    // draw level
    for (int ty = tileVisTop; ty <= tileVisBottom; ++ty)
    {
        for (int tx = tileVisLeft; tx <= tileVisRight; ++tx)
        {
            Tile &t = mWorld[ty * mLevelWidth + tx];
            float x = tx - cx;
            float y = ty - cy;
            core::dimension2du rs((int)scale, (int)scale);
            switch (t.type)
            {
            case TT_OPEN:
                driver->draw2DRectangle(video::SColor(0xff88eeff), core::recti(core::vector2di((int)(x * scale + scx), (int)(y * scale + scy)), rs));
                break;
            case TT_WALL:
                driver->draw2DRectangle(video::SColor(0xff444444), core::recti(core::vector2di((int)(x * scale + scx), (int)(y * scale + scy)), rs));
                break;
            }
            // draw red rect of plr tile
            //if ((int)floorf(mPlayer.y) == ty && tx == mPlayer.hx)
            //    driver->draw2DRectangle(video::SColor(0xffff0000), core::recti(core::vector2di((int)(x * scale + scx), (int)(y * scale + scy)), rs));
        }
    }
    g->beginRender(); //TMP - render() will call endRender so we need to begin one again + drawing below. (we use some base irr drawrects above temporarily so we needed to end render for that)

    // draw player
    {
        //const core::vector2di pos((int)((mPlayer.x - cx) * scale - scale * 0.4f + scx), (int)((mPlayer.y - cy) * scale - scale * 0.4f + scy));
        //const core::recti rect(pos, core::dimension2du((int)(scale * 0.8f), (int)(scale * 0.8f)));
        //driver->draw2DRectangle(video::SColor(0xff88ccaa), rect);
        float px = (mPlayer.x - cx) * scale + scx;
        float py = (mPlayer.y - cy) * scale + scy;
        int sprIdx = 8;
        drawSprite(px, py, sprIdx, false);
    }
*/

    mStages.render();

    STAGE_PLR_TYPE stagePlrType = mStages.getStagePlayerType();
    if (stagePlrType != STAGE_PLR_NONE)
    {
        int sprIdx = mPlayer.spriteDir * 4;
        if (mPlayer.state == PS_IDLE)
        {
            int blinkFreq = (mStages.getCurrentStage() == STAGE_SCHOOL) ? 13 : 25; // blink more frequently as a child
            const bool blink = (((mGameLogicTime / 200) * 16777619) % blinkFreq == 0);
            if (blink)
                sprIdx += 3;
        }
        else if (mPlayer.state == PS_MOVING)
        {
            const int sprCycle[4] = { 0, 1, 0, 2 };
            sprIdx += sprCycle[(mGameLogicTime / 250) % 4];
        }
        drawPlayerSprite(mPlayer.x, mPlayer.y, stagePlrType, sprIdx);
    }

    // draw score (hearts)
    float scoreW = 0;
    if (mScore > 0 && mStages.getCurrentStage() != STAGE_BABY && !mEndingUIActive) // hack - don't show hearts at all in baby stage, so that attention isn't diverted to corner of screen if you get 1st one
    {
        char buf[100];
        sprintf(buf, "%d", mScore);
        //mFont->drawString(40, 40, "score", IF_ALIGN_TOP_LEFT, 0.7f, Color(0x88000000));
        mImages[IMG_HEART]->draw(8, HEIGHT - 12);
        mFont->drawString(22, HEIGHT - 0, buf, IF_ALIGN_BOTTOM_LEFT, 1, Color(0x80aa0000));
        mFont->drawString(22, HEIGHT - 1, buf, IF_ALIGN_BOTTOM_LEFT, 1, Color(0xffee8888));
        scoreW = 40.f;
    }

    // draw ctrl help
    if (!mEndingUIActive)
    {
        mFont->drawString(scoreW + 4, HEIGHT - 0, mStages.getStagePhaseName(), IF_ALIGN_BOTTOM_LEFT, 1, Color(0x80000000));
        mFont->drawString(scoreW + 4, HEIGHT - 1, mStages.getStagePhaseName(), IF_ALIGN_BOTTOM_LEFT, 1, Color(0x80ffffff));
        mFont->drawString(WIDTH - 4, HEIGHT - 0, mStages.getStageTitle(), IF_ALIGN_BOTTOM_RIGHT, 1, Color(0xaa000000));
        mFont->drawString(WIDTH - 4, HEIGHT - 1, mStages.getStageTitle(), IF_ALIGN_BOTTOM_RIGHT, 1, Color(0xaaffffff));
        mFont->drawString(WIDTH / 2, HEIGHT - 0, mStages.getStageCtrlHelpString(), IF_ALIGN_BOTTOM_HCENTER, 1, Color(0xff000000));
        mFont->drawString(WIDTH / 2, HEIGHT - 1, mStages.getStageCtrlHelpString(), IF_ALIGN_BOTTOM_HCENTER, 1);
    }

    if (mEndingUIActive)
    {
        mFont->drawString(360, 50, "Thanks for playing LIFELONE.", IF_ALIGN_CENTER, 1, Color(0xffccddcc));
        mFont->drawText(360, 100, "Created by Jetro Lauha\nfor 22th Ludum Dare 48h game compo\nDec 2011", IF_ALIGN_CENTER, 1, 0, Color(0xff666666));

        static int endUIStartTime = 0;
        if (endUIStartTime == 0)
            endUIStartTime = mStateTime;

        UI8 heartFade = (UI8)(sBoxStep((float)(mStateTime - endUIStartTime), 0, 1000) * 255);
        UI8 escFade = (UI8)(sBoxStep((float)(mStateTime - endUIStartTime), 1000, 2000) * 255);

        mFont->drawText(380, 280, "You got:", IF_ALIGN_TOP_RIGHT, 1, 0, Color(255, 255, 255, heartFade));
        char buf[100];
        sprintf(buf, "%d", mScore);
        mImages[IMG_HEART]->draw(410, 286);
        mFont->drawString(390, 280, buf, IF_ALIGN_TOP_LEFT, 1, Color(0xaa, 0, 0, heartFade));
        mFont->drawString(390, 280, buf, IF_ALIGN_TOP_LEFT, 1, Color(0xee, 0x88, 0x88, heartFade));
        if (mScore == 0)
            mFont->drawString(360, 310, "Play again to find hearts.", IF_ALIGN_TOP_HCENTER, 1, Color(0xaa, 0x88, 0x88, heartFade));
        else if (mScore == 1)
            mFont->drawString(360, 310, "Play again to find more.", IF_ALIGN_TOP_HCENTER, 1, Color(0xbb, 0x88, 0x88, heartFade));
        else if (mScore == 2)
            mFont->drawString(360, 310, "That's all you there is!", IF_ALIGN_TOP_HCENTER, 1, Color(0xcc, 0x99, 0x99, heartFade));

        mFont->drawString(360, 400, "Now press Esc to quit.", IF_ALIGN_CENTER, 1, Color(0x88, 0xdd, 0xdd, escFade));
    }

    g->flush();
    if (mFadePhase != 0.0f)
    {
        float fade = core::clamp<float>(fabsf(mFadePhase), 0, 1);
        gIrrDrawFullscreenQuad(mVideo, video::SColor((irr::u32)(fade * 255), 0, 0, 0), 0, 0);
    }
}
